/*
 * Written by Jan Rekorajski <baggins@mimuw.edu.pl> 1999/08/08
 * 	based od pam_cracklib by Cristian Gafton <gafton@redhat.com>
 * Password generator code by Tom Van Vleck <thvv@multicians.org>
 */

#include "config.h"
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <stdarg.h>
#include <math.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <ctype.h>
#include <syslog.h>

#ifndef DEFAULT_TRIGRAM_FILE
#define	DEFAULT_TRIGRAM_FILE	"/etc/security/trigram"
#endif

#define		PWGEN_USAGE	\
"RECOGNIZED ARGUMENTS:\n\n"							\
"\tminlen=N	minimal length of generated password. Default N=10.\n\n"	\
"\tcount=N		propose N passwords for the user. Default N=5.\n\n"	\
"\ttrifile=path	path to file with letter statistics used by\n"			\
"\t		pronouncable password generator.\n"				\
"\t		Default path=/etc/security/trigram.\n\n"			\
"\tupper=N		Uppercase N letters. Default N=0, max N=minlen/2.\n\n"	\
"\tdigit=N		Convert N letters to digits, module tries to be\n"	\
"\t		smart and first changes look-alike letters\n"			\
"\t		(A=4,E=3,I=1,O=0,S=5,T=7,Z=2).\n"				\
"\t		Default N=0, max N=minlen/2.\n"

#define PWGEN_PROMPT1	"Do you want (choose one letter only):\n\n"\
			"	Pronounceable passwords generated for you (g) ?\n"\
			"	A string of letters/numbers generated for you (l) ?\n"\
			"	A string of characters generated for you (c) ?\n\n"\
			"Enter choice here (q to quit): "

#define PWGEN_PROMPT2	"\nGenerating %d random %s.\n"\
			"	(Password generation may be a bit slow.)\n"\
			"	Hit <RETURN> or <ENTER> until you like the choice.\n"\
			"	Use Ctrl-C to exit.\n\n"

#define GENSTR01	"pronounceable passwords"
#define GENSTR02	"strings of characters"
#define GENSTR03	"strings of letters/numbers"

#define GENTYPE_NONE	0
#define GENTYPE01	1
#define GENTYPE02	2
#define GENTYPE03	3

struct tris_data {
	long sigma;		/* Total letters */
	int tris[26][26][26];	/* Trigraph frequencies */
};

struct gpassword {
	int length;
	char passwd[128];
};

static char *gen_pronouncable_passwd(int length, void *tr);
static char *gen_random_passwd(int length, void *null_data);

static int select_generator(void)
{
	int pwgen_func = GENTYPE_NONE;
	char buf[8];

	printf(PWGEN_PROMPT1);
	fflush(stdout);
	read(0, buf, 8);
	switch (tolower(buf[0])) {
		case 'g':
			pwgen_func = GENTYPE01;
			break;
		case 'c':
			pwgen_func = GENTYPE02;
			break;
		case 'l':
			pwgen_func = GENTYPE03;
			break;
		case 'q':
			pwgen_func = GENTYPE_NONE;
		default:
			break;
	}

	return pwgen_func;
}

/* Main generator */
/* GPW - Generate pronounceable passwords
   This program uses statistics on the frequency of three-letter sequences
   in your dictionary to generate passwords.  The statistics are in trigram.h,
   generated there by the program loadtris.  Use different dictionaries
   and you'll get different statistics.

   This program can generate every word in the dictionary, and a lot of
   non-words.  It won't generate a bunch of other non-words, call them the
   unpronounceable ones, containing letter combinations found nowhere in
   the dictionary.  My rough estimate is that if there are 10^6 words,
   then there are about 10^9 pronounceables, out of a total population
   of 10^11 8-character strings.  I base this on running the program a lot
   and looking for real words in its output.. they are very rare, on the
   order of one in a thousand.

   This program uses "drand48()" to get random numbers, and "srand48()"
   to set the seed to the microsecond part of the system clock.  Works
   for AIX C++ compiler and runtime.  Might have to change this to port
   to another environment.

   The best way to use this program is to generate multiple words.  Then
   pick one you like and transform it with punctuation, capitalization,
   and other changes to come up with a new password.

   THVV 6/1/94 Coded

 */

static int read_trigrams(char *name, struct tris_data *tr)
{
	FILE *f;
	int c1, c2, c3;

	f = fopen(name, "rt");
	if (f == NULL)
		return 1;

	fscanf(f, "%ld\n", &tr->sigma);
/*	D(("%ld", tr->sigma));*/

	for (c1 = 0; c1 < 26; c1++) {
		for (c2 = 0; c2 < 26; c2++) {
			fscanf(f, "\n%*s");
			for (c3 = 0; c3 < 26; c3++) {
				tr->tris[c1][c2][c3] = 0;
				fscanf(f, "%d", &tr->tris[c1][c2][c3]);
/*				D(("%d", tr->tris[c1][c2][c3]));*/
			}
		}
	}

	fclose(f);
	return 0;
}

static char *gen_pronouncable_passwd(int length, void *tr)
{
	int c1, c2, c3;		/* array indices */
	long sumfreq;		/* total frequencies[c1][c2][*] */
	double pik;		/* raw random number in [0.1] from drand48() */
	long ranno;		/* random number in [0,sumfreq] */
	long sum;		/* running total of frequencies */
	int nchar;		/* number of chars in password so far */
	static char password[128];	/* buffer to develop a password */
	struct tris_data *tris = tr;

	/* Pick a random starting point. */
	/* (This cheats a little; the statistics for three-letter
	   combinations beginning a word are different from the stats
	   for the general population.  For example, this code happily
	   generates "mmitify" even though no word in my dictionary
	   begins with mmi. So what.) */

	pik = drand48();	/* random number [0,1] */
	sumfreq = tris->sigma;	/* sigma calculated by loadtris */
	ranno = pik * sumfreq;	/* Weight by sum of frequencies. */
	sum = 0;

	for (c1 = 0; c1 < 26; c1++) {
		for (c2 = 0; c2 < 26; c2++) {
			for (c3 = 0; c3 < 26; c3++) {
				sum += tris->tris[c1][c2][c3];
				if (sum > ranno) {	/* Pick first value */
					password[0] = 'a' + c1;
					password[1] = 'a' + c2;
					password[2] = 'a' + c3;
					c1 = c2 = c3 = 26;	/* Break all loops. */
				}
			}
		}
	}

	/* Do a random walk. */
	nchar = 3;		/* We have three chars so far. */
	while (nchar < length) {
		password[nchar] = '\0';
		password[nchar + 1] = '\0';
		c1 = tolower(password[nchar - 2]) - 'a';	/* Take the last 2 chars */
		c2 = tolower(password[nchar - 1]) - 'a';	/* .. and find the next one. */
		sumfreq = 0;
		for (c3 = 0; c3 < 26; c3++)
			sumfreq += tris->tris[c1][c2][c3];
		/* Note that sum < duos[c1][c2] because
		   duos counts all digraphs, not just those
		   in a trigraph. We want sum. */
		if (sumfreq == 0) {	/* If there is no possible extension.. */
			break;	/* Break while nchar loop & print what we have. */
		}
		/* Choose a continuation. */
		pik = drand48();
		ranno = pik * sumfreq;	/* Weight by sum of frequencies for row. */
		sum = 0;
		for (c3 = 0; c3 < 26; c3++) {
			sum += tris->tris[c1][c2][c3];
			if (sum > ranno) {
				password[nchar++] = 'a' + c3;
				c3 = 26;	/* Break the for c3 loop. */
			}
		}
	}
	return password;
}

static char *gen_random_passwd(int length, void *null_data UNUSED)
{
	static const char ASCII[94] = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ0123456789`~!@#$%^&*()_+-=[]{};':\",./<>?\\|";
	static char password[128];	/* buffer to develop a password */
	int i;
	char c;

	for (i=0; i<length; i++) {
		c = ASCII[(int)(drand48()*94)];
		password[i] = c;
	}
	password[i] = '\0';

	return password;
}

static char *gen_random_passwd1(int length, void *null_data UNUSED)
{
	static const char LD[36] = "abcdefghijklmnopqrstuvwxyz0123456789";
	static char password[128];	/* buffer to develop a password */
	int i;
	char c;

	for (i=0; i<length; i++) {
		c = LD[(int)(drand48()*36)];
		password[i] = c;
	}
	password[i] = '\0';

	return password;
}

static void complicate(int upper, int digit, char *buf)
{
	int i, isdig;
	unsigned int pos, len;

	len = strlen(buf);
	isdig = strcspn(buf, "aeiostz") == len ? 1 : 0;
	if (isdig) {
		do {
			pos = strcspn(buf, "aeiostz");
			if (pos == len)
				break;
			switch (buf[pos]) {
				case 'a':
					buf[pos] = '4';
					break;
				case 'e':
					buf[pos] = '3';
					break;
				case 'i':
					buf[pos] = '1';
					break;
				case 'o':
					buf[pos] = '0';
					break;
				case 's':
					buf[pos] = '5';
					break;
				case 't':
					buf[pos] = '7';
					break;
				case 'z':
					buf[pos] = '2';
					break;
				default:
					buf[pos] = '0' + (int)(drand48()*10);
					break;
			}
		} while (--digit);
	}

	for (i=0; i<digit; i++) {
		pos = (int)(drand48()*len);
		if (!isdigit(buf[pos]))
			buf[pos] = '0' + (int)(drand48()*10);
		else
			i--;
	}

	for (i=0; i<upper; i++) {
		pos = (int)(drand48()*len);
		if (!isdigit(buf[pos]))
			buf[pos] = toupper(buf[pos]);
		else
			i--;
	}

}

int main(int argc, const char **argv)
{
	char tri_file[128];
	int min_length = 10;
	int pw_count = 5;
	int upper = 0;
	int digit = 0;
	const char *gen_prompt;
	char * (*genfn)(int, void *) = NULL;
	int genpw;
	struct tris_data *tr;
	struct stat st;
	char **av;

	strcpy(tri_file, DEFAULT_TRIGRAM_FILE);

	/* step through arguments */
	for (av=argv+1; *av; av++) {
		char *ep = NULL;

		/* generic options */

		if (!strncmp(*av, "minlen=", 7)) {
			min_length = strtol(*av + 7, &ep, 10);
			if (!ep || (min_length < 5))
				min_length = 5;
			if (min_length > 64)
				min_length = 64;
		} else if (!strncmp(*av, "count=", 6)) {
			pw_count = strtol(*av + 6, &ep, 10);
			if (!ep || (pw_count < 5))
				pw_count = 5;
			if (pw_count > 100)
				pw_count = 100;
		} else if (!strncmp(*av, "trifile=", 8)) {
			strncpy(tri_file, *av + 8, 128);
		} else if (!strncmp(*av, "upper=", 6)) {
			upper = strtol(*av + 6, &ep, 10);
			if (!ep || (upper < 0))
				upper = 0;
		} else if (!strncmp(*av, "digit=", 6)) {
			digit = strtol(*av + 6, &ep, 10);
			if (!ep || (digit < 0))
				digit = 0;
		} else {
			fprintf(stderr, PWGEN_USAGE);
			return 1;
		}
	}
	if (upper > min_length/2)
		upper = min_length/2;
	if (digit > min_length/2)
		digit = min_length/2;

	if ((stat(tri_file, &st) == -1) && st.st_size) {
		fprintf(stderr, "trigram path '%s' is invalid\n", tri_file);
		return 1;
	}
	/*
	 * Planned modus operandi:
	 * 1) Present the generators.
	 * 2) Get a response.
	 * 3) Present passwords or return PAM_IGNORE.
	 * 4) Get a passwd.
	 * 5) If okay get it a second time.
	 *      else goto 3
	 * 6) Check to be the same with the first one.
	 * 7) set PAM_AUTHTOK and return
	 */

	srand48(time(NULL));	/* Set random seed. */
	
	genpw = select_generator();

	switch (genpw) {
		case GENTYPE01:
			genfn = gen_pronouncable_passwd;
			gen_prompt = GENSTR01;
			tr = malloc(sizeof(struct tris_data));
			if (tr == NULL)
				return 1;
			if (read_trigrams(tri_file, tr) < 0)
				return 1;
			break;
		case GENTYPE02:
			genfn = gen_random_passwd;
			gen_prompt = GENSTR02;
			tr = NULL;
			break;
		case GENTYPE03:
			genfn = gen_random_passwd1;
			gen_prompt = GENSTR03;
			tr = NULL;
			break;
		case GENTYPE_NONE:
		default:
			return 0;
	}

	do {
		char *pw_buff;
		int i;

		printf(PWGEN_PROMPT2, pw_count, gen_prompt);

		i = 0;
		do {
			pw_buff = genfn(min_length + (rand() & 7), tr);
			if (genpw == GENTYPE01)
				complicate(upper, digit, pw_buff);
			if (genpw == GENTYPE03)
				complicate(upper, 0, pw_buff);
			printf("%s\n", pw_buff);
		} while (++i < pw_count);

		read(0, &i, 1);
	} while (1);

	/* Not reached */
	return 1;
}
